"""Utilities for saving structures to disk."""
import os
import csv

import numpy as np

from skm_pyutils.py_path import make_path_if_not_exists
from skm_pyutils.py_log import log_exception

def val_to_str(val):
    if isinstance(val, str):
        out_str = val.replace(" ", "_")
    elif isinstance(val, float):
        out_str = "{:4f}".format(val)
    elif isinstance(val, int):
        out_str = str(val)
    else:
        str_val = str(val).replace("\n", " ").replace(",", "")
        out_str = '"{}"'.format(str_val)
    return out_str

def arr_to_str(name, arr):
    out_str = name
    for val in arr:
        str_val = val_to_str(val)
        out_str = "{}, {}".format(out_str, str_val)

    return out_str


def save_mixed_dict_to_csv(in_dict, out_dir, out_name="results.csv", save=True):
    """
    Save a dictionary with mixed value types to a csv.

    Currently dict, np.ndarray, and list are supported values.
    Each key in the dictionary is saved as a row in the output csv.

    Args:
        in_dict (dict): The dictionary to save to a csv.
        out_dir (str): The directory to save the csv to.
        out_name (str, optional): Defaults to "results.csv".
        save (bool, optional): Defaults to True.

    Returns:
        The string representation of the data saved to csv.

    """
    full_str = ""
    if save:
        out_loc = os.path.join(out_dir, out_name)
        make_path_if_not_exists(out_loc)
        print("Saving mixed dict data to {}".format(out_loc))
        f = open(out_loc, "w")

    for key, val in in_dict.items():
        if isinstance(val, dict):
            if len(val) == 0:
                out_str = "{},Empty".format(key)
            else:
                out_str = ""
                for k2, val2 in val.items():
                    name = "{} -- {}".format(key, k2)
                    out_str += arr_to_str(name, val2) + "\n"
                out_str = out_str[:-1]
        elif isinstance(val, np.ndarray):
            out_str = arr_to_str(key, val.flatten())
        elif isinstance(val, list):
            out_str = arr_to_str(key, val)
        else:
            out_str = "{},{}".format(key, val_to_str(val))
        full_str += out_str + "\n"
        if save:
            f.write(out_str + "\n")
    
    if save:
        f.close()

    return full_str


def save_dicts_to_csv(filename, in_dicts, do_sort=True):
    """
    Save a list of dictionaries to a csv, cols=vals, rows=dicts.

    The headers are set as the maximal set of keys in in_dicts.
    It is assumed that all other dicts will have a subset of these keys.
    Each entry in the dict is saved to a row of the csv, so it is assumed
    the values in the dict are mostly floats / ints / etc.

    Parameters
    ----------
    filename : str
        The name of the csv file to save results to.
    in_dicts : List
        A list of dictionaries to save to csv.
    do_sort : bool, optional
        Whether to sort the keys after appending, by default True.

    Returns
    -------
    None

    """
    # first, find the dict with the most keys
    if len(in_dicts) == 0:
        return
    max_key = list(in_dicts[0].keys())
    for in_dict in in_dicts:
        names = in_dict.keys()
        if len(names) > len(max_key):
            max_key = list(names)

    # Then append other keys if still missing keys
    did_append = False
    for in_dict in in_dicts:
        names = in_dict.keys()
        for name in names:
            if name not in max_key:
                did_append = True
                max_key.append(name)
    if did_append and do_sort:
        max_key = sorted(max_key)
    max_key_friendly = [k.replace(" ", "_") for k in max_key]

    try:
        print("Saving summary data to {}".format(filename))
        make_path_if_not_exists(filename)
        with open(filename, "w", newline="") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=max_key)
            writer.writerow(dict(zip(max_key, max_key_friendly)))
            for in_dict in in_dicts:
                writer.writerow(in_dict)

    except Exception as e:
        log_exception(e, "When {} saving to csv".format(filename))



"""Utilities for plotting with matplotlib."""

import colorsys
import os
from collections import OrderedDict

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import seaborn as sns

from skm_pyutils.py_path import make_path_if_not_exists


class GroupManager:
    def __init__(self, group_list):
        self.group_list = group_list
        print(self.group_list)
        self.info_dict = OrderedDict()
        self.index = 0
        self.color_list = ["Blues", "Oranges", "Greens", "Reds", "Purples", "Greys"]
        set_vals = sorted(set(group_list), key=group_list.index)
        print(set_vals)
        import numpy as np

        if len(set_vals) > len(self.color_list):
            start_vals = np.arange(0.0, 2.5, 2.5 / (len(set_vals) - 0.99))
            for set_v, start_v in zip(set_vals, start_vals):
                self.info_dict[set_v] = ColorManager(
                    group_list.count(set_v), "sns_helix", start=start_v
                )
        else:
            start_vals = self.color_list[: len(set_vals)]
            for set_v, start_v in zip(set_vals, start_vals):
                self.info_dict[set_v] = ColorManager(
                    group_list.count(set_v), "sns", sns_style=start_v
                )
        print(self.info_dict)

    def get_next_color(self):
        out = self.info_dict[self.group_list[self.index]].get_next_color()
        self._increment()
        return out

    def _increment(self):
        self.index += 1
        if self.index == len(self.group_list):
            self.index = 0

    def test_plot(self):
        import numpy as np
        from scipy.stats import norm

        fig, ax = plt.subplots()
        x_axis = np.arange(-15, 5, 0.001)
        std_devs = np.arange(0.8, 3, 2.20 / len(self.group_list))
        for sd in std_devs:
            ax.plot(x_axis, norm.pdf(x_axis, -sd * 2, sd), color=self.get_next_color())
        sns.despine(top=True, bottom=True, right=True, left=True)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        fig.savefig("test.png", dpi=400)


class ColorManager:
    def __init__(self, num_colors, method="sns", **kwargs):
        self.num_colors = num_colors
        if method == "sns":
            sns_style = kwargs.get("sns_style", None)
            self.create_sns_palette(s_type=sns_style)
        elif method == "rgb":
            self.create_rgb()
        elif method == "sns_helix":
            start = kwargs.get("start", 0)
            self.create_sns_helix(start)
        else:
            raise ValueError("{} not recognised method".format(method))
        self.idx = 0

    def create_rgb(self):
        N = self.num_colors
        HSV_tuples = [(x * 1.0 / N, 0.5, 0.5) for x in range(N)]
        RGB_tuples = map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples)
        self.colors = list(RGB_tuples)

    def create_sns_palette(self, s_type=None):
        self.colors = sns.color_palette(s_type, self.num_colors)

    def create_sns_helix(self, start):
        self.colors = sns.cubehelix_palette(self.num_colors, start)

    def get_next_color(self):
        result = self.colors[self.idx]
        self._increment()
        return result

    def get_color(self, index):
        return self.colors[index]

    def test_plot(self):
        import numpy as np
        from scipy.stats import norm

        fig, ax = plt.subplots()
        x_axis = np.arange(-15, 5, 0.001)
        std_devs = np.arange(0.8, 3, 2.20 / self.num_colors)
        for sd in std_devs:
            ax.plot(x_axis, norm.pdf(x_axis, -sd * 2, sd), color=self.get_next_color())
        sns.despine(top=True, bottom=True, right=True, left=True)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        fig.savefig("test.png", dpi=400)

    def _increment(self):
        """Private function to increase the internal idx counter."""
        self.idx = self.idx + 1
        if self.idx == self.num_colors:
            self.idx = 0


class GridFig:
    """Handles gridded figures."""

    def __init__(
        self,
        rows,
        cols=None,
        size_multiplier_x=5,
        size_multiplier_y=5,
        wspace=0.3,
        hspace=0.3,
        tight_layout=False,
        traverse_rows=True,
    ):
        """
        Set up the grid specifications.
    
        If only rows is passed, the number of columns and rows
        are automatically determined to be close to a square.

        size_multiplier, wspace, and hspace are used for spacing

        """
        if cols is None:
            rows, cols = self.auto_determine_grid(rows)
        self.fig = plt.figure(
            figsize=(cols * size_multiplier_x, rows * size_multiplier_y),
            tight_layout=tight_layout,
        )
        self.gs = gridspec.GridSpec(rows, cols, wspace=wspace, hspace=hspace)
        self.idx = 0
        self.rows = rows
        self.cols = cols
        self.along_rows = traverse_rows

    def auto_determine_grid(self, rows):
        closest_sqrt = 1
        while (closest_sqrt * closest_sqrt) < rows:
            closest_sqrt += 1
        return closest_sqrt, closest_sqrt

    def get_ax(self, row_idx, col_idx, circular=False):
        """Add subplot with standard 1x1 gs -> returns ax."""
        if circular:
            ax = self.fig.add_subplot(self.gs[row_idx, col_idx], projection="polar")
        else:
            ax = self.fig.add_subplot(self.gs[row_idx, col_idx])
        return ax

    def get_multi_ax(self, row_start, row_end, col_start, col_end):
        """Add subplot with custom gs sizes -> returns ax."""
        ax = self.fig.add_subplot(self.gs[row_start:row_end, col_start:col_end])
        plt.subplots_adjust(top=0.85)
        return ax

    def save_fig(self, out_dir, out_name):
        """Names and saves figure."""
        out_loc = os.path.join(out_dir, out_name)
        print("Saved figure to {}".format(out_loc))
        make_path_if_not_exists(out_loc)
        self.fig.savefig(out_loc, dpi=400)
        plt.close(self.fig)

    def savefig(self, fname, **kwargs):
        """Passes all to matplotlib savefig call"""
        print("Saved figure to {}".format(fname))
        make_path_if_not_exists(fname)
        if "dpi" not in kwargs.keys():
            kwargs["dpi"] = 400
        self.fig.savefig(fname, **kwargs)
        plt.close(self.fig)

    def get_fig(self):
        """Return the figure object in this class."""
        return self.fig

    def get_next(self, circular=False):
        """
        Get next index along rows or columns.

        along rows:
        1   2   3   4   5   6
        7   8   9   10  11  12  ...

        else:
        1   3   5   7   9   11
        2   4   6   8   10  12  ...

        """
        if self.along_rows:
            row_idx = self.idx // self.cols
            col_idx = self.idx % self.cols

            ax = self.get_ax(row_idx, col_idx, circular=circular)
        else:
            row_idx = self.idx % self.rows
            col_idx = self.idx // self.rows

            ax = self.get_ax(row_idx, col_idx, circular=circular)
        self._increment()
        return ax

    def get_next_snake(self):
        """
        Get the next index in a snake like pattern.

        1   2   5   6   9   10  ...
        3   4   7   8   11  12  ..

        """
        if self.rows != 2:
            raise ValueError("Can't get snake unless there are two rows")
        i = self.idx
        row_idx = (i // 2) % 2
        col_idx = (i // 2) + (i % 2) - row_idx
        ax = self.get_ax(row_idx, col_idx)
        self._increment()
        return ax

    def _increment(self):
        """Private function to increase the internal idx counter."""
        self.idx = self.idx + 1
        if self.idx == self.rows * self.cols:
            self.idx = 0


class UnicodeGrabber(object):
    """This is a fully static class to get unicode chars for plotting."""

    char_dict = {
        "micro": "\u00B5",
        "pow2": "\u00B2",
    }

    @staticmethod
    def get_chars():
        return list(UnicodeGrabber.char_dict.keys())

    @staticmethod
    def get(char, default=""):
        return UnicodeGrabber.char_dict.get(char, default)

    @staticmethod
    def to_sub(num):
        return chr(int(f"208{num}", 16))



"""Utilities for pandas dataframes."""
import os

import pandas as pd


def list_to_df(in_list, transpose=False, headers=None):
    """
    Convert a list to a dataframe with the given headers.

    Tranpose handles the list shape.
    If transpose is False, list is assumed to be like
    [
        (row 1) [1_1, 1_2, ..., 1_N]
        ...
        (row M) [M_1, M_2, ..., M_N]
    ]
    Otherwise, the list is assumed to be like
    [
        (col 1) [1_1, 2_1, ..., M_1]
        ...
        (col N) [1_N, 2_N, ..., M_N]
    ]
    
    Parameters
    ----------
    in_list : list
        The list to convert.
    tranpose : bool, optional
        Whether to transpose the list, by default False.
    headers : list, optional
        A list of headers for the data. By default is V1, V2, ... VN.

    Returns
    -------
    pandas.DataFrame

    """
    if headers is None:
        if not transpose:
            headers = ["V{}".format(i) for i in range(len(in_list[0]))]
        else:
            headers = ["V{}".format(i) for i in range(len(in_list))]

    if transpose:
        df = pd.DataFrame(in_list).T
        df.columns = headers
    else:
        df = pd.DataFrame.from_records(in_list, columns=headers)
    return df


def df_from_file(filename):
    """
    Read a pandas.DataFrame from filename.

    Parameters
    ----------
    filename : str
        The path to the file to read

    Returns
    -------
    pandas.DataFrame or TextParser
        The read data

    """
    ext = os.path.splitext(filename)[1]
    if ext == ".psv":
        df = pd.read_csv(filename, delimiter="|")
    elif ext == ".csv":
        df = pd.read_csv(filename)
    elif ext == ".xlsx":
        df = pd.read_excel(filename)
    else:
        raise ValueError(f"Unsupported file extension {ext}")
    return df


def df_to_file(df, filename, index=False, **kwargs):
    """
    Save a pandas.DataFrame to filename.

    Parameters
    ----------
    df : pandas.DataFrame
        The input dataframe to save.
    filename : str
        The path of the file to save to.
    index : bool
        Whether to write row names, by default False.
    kwargs : keyword arguments
        Passed to pandas method.

    Returns
    -------
    None

    """

    def get_to_retry():
        print("{} may currently be in use, try closing it".format(filename))
        retry = True
        continue_ = True
        while retry:
            done = input("When closed, please enter y to retry, or q to quit:\n")
            if len(done) == 0:
                retry = True
            elif done.strip().lower() == "y":
                retry = False
            elif done.strip().lower() == "q":
                retry = False
                continue_ = False
        if continue_:
            print("Retrying saving to {}".format(filename))
        return continue_

    ext = os.path.splitext(filename)[1]
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    if ext == ".psv":
        try:
            df.to_csv(filename, sep="|", index=index, **kwargs)
        except PermissionError:
            continue_ = get_to_retry()
            if continue_:
                df.to_csv(filename, sep="|", index=index, **kwargs)
    elif ext == ".csv":
        try:
            df.to_csv(filename, sep=",", index=index, **kwargs)
        except PermissionError:
            continue_ = get_to_retry()
            if continue_:
                df.to_csv(filename, sep=",", index=index, **kwargs)
    elif ext == ".xlsx":
        try:
            df.to_excel(filename, index=index, **kwargs)
        except PermissionError:
            continue_ = get_to_retry()
            if continue_:
                df.to_excel(filename, index=index, **kwargs)
    else:
        raise ValueError(f"Unsupported file extension {ext}")


def df_subset_from_rows(df, rows):
    """
    Get a subset of a dataframe based on row indices.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to subset from.
    rows : list-like of int
        The rows to grab.
    
    Returns
    -------
    pandas.DataFrame

    """
    df_subset = df.iloc[rows].copy().reset_index()
    return df_subset


def show_interactive_table(table, notebook=False) -> None:
    import dtale

    ## TODO maybe should have a config for notebook version
    if notebook:
        dtale.show(table)
    else:
        dtale.show(table).open_browser()



"""Utilities for code performance profiling."""
import os
import io
import argparse
import pstats
import cProfile

def profileit(name):
    """
    Decorator to profile code performance.

    Usage:
    @profileit("profile_for_func1_001");
    def func1(...)

    See https://stackoverflow.com/questions/5375624/a-decorator-that-profiles-a-method-call-and-logs-the-profiling-result/5376616

    """
    def inner(func):
        def wrapper(*args, **kwargs):
            prof = cProfile.Profile()
            retval = prof.runcall(func, *args, **kwargs)
            s = io.StringIO()
            sortby = pstats.SortKey.CUMULATIVE
            ps = pstats.Stats(prof, stream=s).strip_dirs().sort_stats(sortby)
            ps.print_stats()
            with open(name, "w") as f:
                f.write(s.getvalue())
            return retval
        return wrapper
    return inner


def default_stats(profile_location):
    """
    Print the result of code profling.

    See https://docs.python.org/2/library/profile.html
    
    Parameters
    ----------
    profile_location : str
        The path to a profile file.

    Returns
    -------
    None

    """
    if os.path.isfile(profile_location):
        p = pstats.Stats(profile_location)
        # p.strip_dirs().sort_stats(-1).print_stats()

        print("Biggest cumulative users")
        p.sort_stats("cumulative").print_stats(20)

        print("Biggest time users")
        p.sort_stats("time").print_stats(20)

        # p.sort_stats('time', 'cumulative').print_stats(.5, 'init')
        # p.print_callers(.5, 'init')

    else:
        raise FileNotFoundError(
            "{} does not exist".format(profile_location)
            + "You can create profile.out by running:"
            + "python -m cProfile -o <profile_location> <filename>"
        )


def main():
    parser = argparse.ArgumentParser("Command line interface")
    parser.add_argument(
        "profile_location", type=str, help="location of the profile file"
    )
    parsed, unparsed = parser.parse_known_args()

    if len(unparsed) > 0:
        raise ValueError("Unrecognised command line arguments")

    default_stats(parser.profile_location)


if __name__ == "__main__":
    main()



"""Statistics functions and paper reporting."""

import numpy as np
import pingouin
import matplotlib.pyplot as plt
import seaborn as sns

from skm_pyutils.py_plot import UnicodeGrabber
from skm_pyutils.py_table import list_to_df

def plot_dists(x, y, ax=None, **fmt_kwargs):
    """Plot distrubtions of two arrays."""
    palette = fmt_kwargs.get("palette", "dark")
    context = fmt_kwargs.get("context", "paper")
    group1_name = fmt_kwargs.get("group1", "1")
    group2_name = fmt_kwargs.get("group2", "2")
    vname = fmt_kwargs.get("value", "values")
    sns.set_palette(palette)
    if context == "paper":
        sns.set_context(
            "paper", font_scale=1.4, rc={"lines.linewidth": 3.2},
        )
    else:
        sns.set_context(context)

    if ax is None:
        figure, ax = plt.subplots()
    else:
        figure = None
    despine = fmt_kwargs.get("despine", True)
    trim = fmt_kwargs.get("trim", True)
    offset = fmt_kwargs.get("offset", None)

    df_list = []
    for val in x:
        df_list.append([val, group1_name])
    for val in y:
        df_list.append([val, group2_name])

    df = list_to_df(df_list, headers=[vname, "group"])
    sns.kdeplot(data=df, x=vname, hue="group", multiple="stack", ax=ax)

    ax.set_xlabel(vname)
    if despine:
        sns.despine(offset=offset, trim=trim)

    return figure

def plot_corr(x, y, ax=None, **fmt_kwargs):
    """Plot correlation between two arrays."""
    value1_name = fmt_kwargs.get("value1", "1")
    value2_name = fmt_kwargs.get("value2", "2")
    palette = fmt_kwargs.get("palette", "dark")
    context = fmt_kwargs.get("context", "paper")
    sns.set_palette(palette)
    if context == "paper":
        sns.set_context(
            "paper", font_scale=1.4, rc={"lines.linewidth": 3.2},
        )
    else:
        sns.set_context(context)

    if ax is None:
        figure, ax = plt.subplots()
    else:
        figure = None
    despine = fmt_kwargs.get("despine", True)
    trim = fmt_kwargs.get("trim", True)
    offset = fmt_kwargs.get("offset", None)

    sns.regplot(x=x, y=y, ax=ax, truncate=False, order=1)

    extents_x = (0.97 * np.min(x), np.max(x) * 1.03)
    extents_y = (0.97 * np.min(y), np.max(y) * 1.03)

    ax.set_xlim(extents_x)
    ax.set_ylim(extents_y)
    ax.set_xlabel(value1_name)
    ax.set_ylabel(value2_name)
    if despine:
        sns.despine(offset=offset, trim=trim)

    return figure

def corr(x, y, fmt_kwargs=None, do_plot=False, ax=None, **kwargs):
    """
    Compute correlation between x and y.

    Based on the method passed as a keyword arg. By default Pearsons.
    Also returns a formatted string representation.

    Parameters
    ----------
    x : array_like
        First set of observations.
    y: array_like
        Second set of observations.
    fmt_kwargs : dict, optional
        A dictionary of kwargs to control the formatting.
        value - the name of the values being tested
        unit - the unit of the values being tested
        group1 - the name of x
        group2 - the name of y
        signif - the significance level (float)
        n_decimals - the number of decimal places to print (int)
        n_pdecimals - the number of decimal places to print for p (int)
        show_quartiles - include quartiles in report (bool)
        do_print - print the report string (bool)
    do_plot : bool, optional
        Whether or not to plot the result.
    ax : axes object, optional
        An axes to plot into.
    **kwargs : keyword arguments
        These are passed to pingouin.corr
        Of particular is method - the correlation to run.

    Returns
    -------
    dict with keys:
        "results" : pd.DataFrame
            The dataframe of results
        "output" : str
            A string to describe the test result for reporting
        "figure" : matplotlib.pyplot.figure or None
            The returned figure plotted into.

    See also
    --------
    pinouin.corr

    """
    if fmt_kwargs is None:
        fmt_kwargs = {}
    group = fmt_kwargs.get("group", "")
    if group != "":
        group = " " + group
    unit_name = fmt_kwargs.get("unit", "")
    if unit_name != "":
        unit_name = " " + unit_name + ", "
    else:
        unit_name = ", "
    value1_name = fmt_kwargs.get("value1", "1")
    value2_name = fmt_kwargs.get("value2", "2")
    signif_level = fmt_kwargs.get("signif", 0.05)
    n_decimals = fmt_kwargs.get("n_decimals", 2)
    n_pdecimals = fmt_kwargs.get("n_pdecimals", 3)
    show_quartiles = fmt_kwargs.get("show_quartiles", True)
    do_print = fmt_kwargs.get("do_print", True)

    sided = kwargs.get("alternative", "two-sided")
    method = kwargs.get("method", "pearson")

    results_df = pingouin.corr(x, y, **kwargs)

    n = results_df["n"].values[0]
    r = np.round(results_df["r"].values[0], n_decimals)
    P = np.round(results_df["p-val"].values[0], n_pdecimals)
    power = np.round(results_df["power"].values[0], n_decimals)
    ci = np.round(np.array(results_df["CI95%"].values[0]), n_decimals)

    lin_reg_df = pingouin.linear_regression(x, y, as_dataframe=False)
    lm_adjr2 = np.round(lin_reg_df["adj_r2"], n_decimals)

    if method == "pearson":
        co_eff_name = "r"
        corr_name = "Pearson"
    elif method == "spearman":
        co_eff_name = "\u03A1"
        corr_name = "spearman"
    elif method == "kendall":
        co_eff_name = "\u03A4"
        corr_name = "Kendall"

    if r < 0:
        type_corr = "negative"
    if r == 0:
        type_corr = "no"
    if r > 0:
        type_corr = "positive"

    if sided == "two-sided":
        tailed = "(two-tailed)"
    else:
        tailed = "(one-tailed)"

    if P < signif_level:
        differ_str = "was significant"
    else:
        differ_str = "was not significant"

    if show_quartiles:
        result_str = f"There was a {type_corr} {corr_name} correlation of {co_eff_name} = {r} [{ci[0]}, {ci[1]}] 95% CI"
    else:
        result_str = (
            f"There was a {type_corr} {corr_name} correlation of {co_eff_name} = {r}"
        )
    relation_str = f" between {value1_name} and {value2_name}{group}"
    stats_str = f"; this {differ_str} (\u0070 = {P}, N = {n}, power = {power} {tailed}"
    pow2 = UnicodeGrabber.get("pow2")
    lin_reg_str = f", linear regression R{pow2} = {lm_adjr2})"

    final_str = result_str + relation_str + stats_str + lin_reg_str

    if do_print:
        print(final_str)

    figure = None
    if do_plot:
        figure = plot_corr(x, y, ax=ax, **fmt_kwargs)

    results = {
        "results": results_df,
        "output": final_str,
        "figure": figure,
    }

    return results


def mwu(x, y, fmt_kwargs=None, do_plot=False, ax=None, **kwargs):
    """
    Compute the Mann-Whitney U Test.

    This is the independent t-test non-parametric version.
    Also returns a formatted string for paper reporting.

    Parameters
    ----------
    x : array_like
        First set of observations.
    y: array_like
        Second set of observations.
    fmt_kwargs : dict, optional
        A dictionary of kwargs to control the formatting.
        value - the name of the values being tested
        unit - the unit of the values being tested
        group1 - the name of x
        group2 - the name of y
        signif - the significance level (float)
        n_decimals - the number of decimal places to print (int)
        n_pdecimals - the number of decimal places to print for p (int)
        show_quartiles - include quartiles in report (bool)
        do_print - print the report string (bool)
    do_plot : bool, optional
        Whether or not to plot the result.
    ax : axes object, optional
        An axes to plot into.
    **kwargs : keyword arguments
        These are passed to pingouin.corr

    Returns
    -------
    dict with keys:
        "results" : pd.DataFrame
            The dataframe of results
        "output" : str
            A string to describe the test result for reporting
        "figure" : matplotlib.pyplot.figure or None
            The returned figure plotted into.

    See also
    --------
    pinouin.mwu
    scipy.stats.mannwhitneyu

    """
    if fmt_kwargs is None:
        fmt_kwargs = {}
    vname = fmt_kwargs.get("value", "values")
    unit_name = fmt_kwargs.get("unit", "")
    if unit_name != "":
        unit_name = " " + unit_name + ", "
    else:
        unit_name = ", "
    group1_name = fmt_kwargs.get("group1", "1")
    group2_name = fmt_kwargs.get("group2", "2")
    signif_level = fmt_kwargs.get("signif", 0.05)
    n_decimals = fmt_kwargs.get("n_decimals", 2)
    n_pdecimals = fmt_kwargs.get("n_pdecimals", 3)
    show_quartiles = fmt_kwargs.get("show_quartiles", True)
    do_print = fmt_kwargs.get("do_print", True)

    sided = kwargs.get("alternative", "two-sided")

    results_df = pingouin.mwu(x, y, **kwargs)
    U = results_df["U-val"].values[0]
    P = np.round(results_df["p-val"].values[0], n_pdecimals)
    cl = np.round(results_df["CLES"].values[0], n_decimals)
    median1 = np.round(np.nanmedian(x), n_decimals)
    lowerq1, higherq1 = np.round(np.nanpercentile(x, [25, 75]), n_decimals)
    lowerq2, higherq2 = np.round(np.nanpercentile(y, [25, 75]), n_decimals)
    median2 = np.round(np.nanmedian(y), n_decimals)

    sample_size1 = len(x)
    sample_size2 = len(y)
    n1 = "n" + UnicodeGrabber.to_sub(1)
    n2 = "n" + UnicodeGrabber.to_sub(2)
    if sample_size1 == sample_size2:
        sample_str = f"{n1} = {n2} = {sample_size1}"
    else:
        sample_str = f"{n1} = {sample_size1}, {n2} = {sample_size2}"

    stats_str = (
        "(Mann-Whitney "
        + "\u0055"
        + f" = {U}, CLES = {cl}, {sample_str}, "
        + "\u0070"
        + f" = {P}"
    )

    if sided == "two-sided":
        stats_str += " two-tailed)."
    else:
        stats_str += " one-tailed)."

    if P < signif_level:
        differ_str = "differed significantly"
    else:
        differ_str = "did not differ significantly"

    if show_quartiles:
        results_str = (
            f"Median [quartiles] {vname} in groups {group1_name} and {group2_name} were "
            + f"{median1} [{lowerq1}, {higherq1}] and {median2} [{lowerq2}, {higherq2}]"
            f"{unit_name}respectively; "
            f"the distributions in the two groups {differ_str} {stats_str}"
        )
    else:
        results_str = (
            f"Median {vname} in groups {group1_name} and {group2_name} were "
            + f"{median1} and {median2}"
            f"{unit_name}respectively; "
            f"the distributions in the two groups {differ_str} {stats_str}"
        )

    if do_print:
        print(results_str)

    figure = None
    if do_plot:
        figure = plot_dists(x, y, ax=ax, **fmt_kwargs)

    results = {
        "results": results_df,
        "output": results_str,
        "figure": figure,
    }

    return results


def wilcoxon(x, y, fmt_kwargs=None, do_plot=False, ax=None, **kwargs):
    """
    Compute the wilcoxon signed-rank test.

    This is the non-parametric paired t-test.
    Also returns a formatted string for paper reporting.

    Parameters
    ----------
    x : array_like
        First set of observations.
    y: array_like
        Second set of observations.
    fmt_kwargs : dict, optional
        A dictionary of kwargs to control the formatting.
        value - the name of the values being tested
        unit - the unit of the values being tested
        group1 - the name of x
        group2 - the name of y
        signif - the significance level (float)
        n_decimals - the number of decimal places to print (int)
        n_pdecimals - the number of decimal places to print for p (int)
        show_quartiles - include quartiles in report (bool)
        do_print - print the report string (bool)
    do_plot : bool, optional
        Whether or not to plot the result.
    ax : axes object, optional
        An axes to plot into.
    **kwargs : keyword arguments
        These are passed to pingouin.corr

    Returns
    -------
    dict with keys:
        "results" : pd.DataFrame
            The dataframe of results
        "output" : str
            A string to describe the test result for reporting
        "figure" : matplotlib.pyplot.figure or None
            The returned figure plotted into.

    See also
    --------
    pinouin.wilcoxon
    scipy.stats.wilcoxon

    """
    if fmt_kwargs is None:
        fmt_kwargs = {}
    vname = fmt_kwargs.get("value", "values")
    unit_name = fmt_kwargs.get("unit", "")
    if unit_name != "":
        unit_name = " " + unit_name + ", "
    else:
        unit_name = ", "
    group1_name = fmt_kwargs.get("group1", "1")
    group2_name = fmt_kwargs.get("group2", "2")
    signif_level = fmt_kwargs.get("signif", 0.05)
    n_decimals = fmt_kwargs.get("n_decimals", 2)
    n_pdecimals = fmt_kwargs.get("n_pdecimals", 3)
    show_quartiles = fmt_kwargs.get("show_quartiles", True)
    do_print = fmt_kwargs.get("do_print", True)

    sided = kwargs.get("alternative", "two-sided")

    results_df = pingouin.wilcoxon(x, y, **kwargs)
    U = results_df["W-val"].values[0]
    P = np.round(results_df["p-val"].values[0], n_pdecimals)
    cl = np.round(results_df["CLES"].values[0], n_decimals)
    median1 = np.round(np.nanmedian(x), n_decimals)
    lowerq1, higherq1 = np.round(np.nanpercentile(x, [25, 75]), n_decimals)
    lowerq2, higherq2 = np.round(np.nanpercentile(y, [25, 75]), n_decimals)
    median2 = np.round(np.nanmedian(y), n_decimals)

    sample_size1 = len(x)
    sample_size2 = len(y)
    n1 = "n" + UnicodeGrabber.to_sub(1)
    n2 = "n" + UnicodeGrabber.to_sub(2)
    if sample_size1 == sample_size2:
        sample_str = f"{n1} = {n2} = {sample_size1}"
    else:
        sample_str = f"{n1} = {sample_size1}, {n2} = {sample_size2}"

    stats_str = (
        "(Wilcoxon signed-rank "
        + "W"
        + f" = {U}, CLES = {cl}, {sample_str}, "
        + "\u0070"
        + f" = {P}"
    )

    if sided == "two-sided":
        stats_str += " two-tailed)."
    else:
        stats_str += " one-tailed)."

    if P < signif_level:
        differ_str = "differed significantly"
    else:
        differ_str = "did not differ significantly"

    if show_quartiles:
        results_str = (
            f"Median [quartiles] {vname} in groups {group1_name} and {group2_name} were "
            + f"{median1} [{lowerq1}, {higherq1}] and {median2} [{lowerq2}, {higherq2}]"
            f"{unit_name}respectively; "
            f"the distributions in the two groups {differ_str} {stats_str}"
        )
    else:
        results_str = (
            f"Median {vname} in groups {group1_name} and {group2_name} were "
            + f"{median1} and {median2}"
            f"{unit_name}respectively; "
            f"the distributions in the two groups {differ_str} {stats_str}"
        )

    if do_print:
        print(results_str)

    figure = None
    if do_plot:
        figure = plot_dists(x, y, ax=ax, **fmt_kwargs)

    results = {
        "results": results_df,
        "output": results_str,
        "figure": figure,
    }

    return results



"""Path related utility functions."""
import os
import re
import argparse
import shutil
from collections import OrderedDict

from skm_pyutils.py_config import parse_args


def make_path_if_not_exists(fname):
    """Make directory structure for given fname if it does not exist."""
    if os.path.dirname(fname) != "":
        os.makedirs(os.path.dirname(fname), exist_ok=True)


def make_dir_if_not_exists(dirname):
    """Make the directory dirname if it does not exist."""
    os.makedirs(dirname, exist_ok=True)


def has_ext(filename, ext, case_sensitive_ext=False):
    """
    Check if the filename ends in the extension.

    Parameters
    ----------
    filename : str
        The name of the file.
    ext : str
        The extension, may have leading dot (e.g txt == .txt).
    case_sensitive_ext: bool, optional. Defaults to False,
        Whether to match the case of the file extension.

    Returns
    -------
    bool
        Indicates if the filename has the extension

    """
    if ext is None:
        return True
    if ext[0] != ".":
        ext = "." + ext
    if case_sensitive_ext:
        return filename[-len(ext) :] == ext
    else:
        return filename[-len(ext) :].lower() == ext.lower()


def get_all_files_in_dir(
    in_dir,
    ext=None,
    return_absolute=True,
    recursive=False,
    verbose=False,
    re_filter=None,
    case_sensitive_ext=False,
):
    """
    Get all files in the directory with the given extensions.

    Parameters
    ----------
    in_dir : str
        The absolute path to the directory
    ext : str, optional. Defaults to None.
        The extension of files to get.
    return_absolute : bool, optional. Defaults to True.
        Whether to return the absolute filename or not.
    recursive: bool, optional. Defaults to False.
        Whether to recurse through directories.
    verbose: bool, optional. Defaults to False.
        Whether to print the files found.
    re_filter: str, optional. Defaults to None
        a regular expression used to filter the results
    case_sensitive_ext: bool, optional. Defaults to False,
        Whether to match the case of the file extension

    Returns
    -------
    List
        A list of filenames with the given parameters.

    """
    if not os.path.isdir(in_dir):
        raise ValueError("Non existant directory " + str(in_dir))

    def match_filter(f):
        if re_filter is None:
            return True
        search_res = re.search(re_filter, f)
        return search_res is not None

    def ok_file(root_dir, f):
        return (
            has_ext(f, ext, case_sensitive_ext=case_sensitive_ext)
            and os.path.isfile(os.path.join(root_dir, f))
            and match_filter(f)
        )

    def convert_to_path(root_dir, f):
        return os.path.abspath(os.path.join(root_dir, f)) if return_absolute else f

    if verbose:
        print("Adding following files from {}".format(in_dir))

    if recursive:
        onlyfiles = []
        for root, _, filenames in os.walk(in_dir):
            start_root = root[: len(in_dir)]

            if len(root) == len(start_root):
                end_root = ""
            else:
                if in_dir.endswith(os.sep):
                    end_root = root[len(in_dir) :]
                else:
                    end_root = root[len(in_dir + os.sep) :]
            for filename in filenames:
                filename = os.path.join(end_root, filename)
                if ok_file(start_root, filename):
                    to_add = convert_to_path(start_root, filename)
                    if verbose:
                        print(to_add)
                    onlyfiles.append(to_add)

    else:
        onlyfiles = [
            convert_to_path(in_dir, f)
            for f in sorted(os.listdir(in_dir))
            if ok_file(in_dir, f)
        ]
        if verbose:
            for f in onlyfiles:
                print(f)

    if verbose:
        print()
    return onlyfiles


def get_dirs_matching_regex(start_dir, re_filters=None, return_absolute=True):
    """
    Recursively get all directories from start_dir that match regex.

    Parameters
    ----------
    start_dir : str
        The path to the directory to start at.
    re_filter : list of str, optional. Defaults to None.
        The list of regular expressions to match.
        Returns all directories if passed as None.

    Returns
    -------
    list
        A list of directories matching the regex.

    """
    if not os.path.isdir(start_dir):
        raise ValueError("Non existant directory " + str(start_dir))

    def match_filter(f):
        if re_filters is None:
            return True
        for re_filter in re_filters:
            search_res = re.search(re_filter, f.replace(os.sep, "/"))
            if search_res is None:
                return False
        return True

    dirs = []
    for root, _, _ in os.walk(start_dir):
        start_root = root[: len(start_dir)]

        if len(root) == len(start_root):
            end_root = ""
        else:
            if start_dir.endswith(os.sep):
                end_root = root[len(start_dir) :]
            else:
                end_root = root[len(start_dir + os.sep) :]

        if match_filter(end_root):
            to_add = root if return_absolute else end_root
            dirs.append(to_add)
    return dirs


def get_base_dir_to_files(
    filenames, start_dir, ext=None, re_filter=None, print_info=True
):
    """
    Get the base directory of a set of filenames.

    If multiple matches are found, returns all matching filenames.

    Parameters
    ----------
    filenames : list of str
        The filenames to search for.
    start_dir : str
        Where to start the search.
    ext : str, optional
        The extension of the filenames, by default None.
    re_filter : str, optional. 
        A regex to use to find files, by defaults None.
    print_info : bool, optional
        Whether to print info of the search, by default True.

    Returns
    -------
    OrderedDict
        Dictionary of matching file base paths.
    set
        A set of files with no matches
    set
        A set of files with multiple matches
    """
    filenames = list(filenames)
    files_to_check = get_all_files_in_dir(
        start_dir, ext=ext, re_filter=re_filter, recursive=True
    )
    found_dict = OrderedDict()

    for f in files_to_check:
        base = os.path.basename(f)
        dirs = os.path.dirname(f)
        if base in filenames:
            if base not in found_dict.keys():
                found_dict[base] = []
            found_dict[base].append(dirs)

    no_match = []
    multi_match = []
    found = list(found_dict.keys())
    num_found = 0
    for v in filenames:
        if v not in found:
            no_match.append(v)
        else:
            num_found += 1
            if len(found_dict[v]) > 1:
                multi_match.append(v)

    found_dict = OrderedDict(
        sorted(found_dict.items(), key=lambda x: filenames.index(x[0]))
    )

    if print_info:
        to_print = "Found {} files out of {}, {} have multiple matches".format(
            num_found, len(filenames), len(set(multi_match))
        )
        print(to_print)

    return found_dict, set(no_match), set(multi_match)


def cli_entry():
    """Command line interface entry point."""
    parser = argparse.ArgumentParser(description="Directory list command line")

    parser.add_argument(
        "directory", type=str, help="Directory to find files from.",
    )
    parser.add_argument(
        "--recursive",
        "-r",
        action="store_true",
        help="Whether to recurse into subdirectories.",
    )
    parser.add_argument(
        "--extension", "-e", type=str, default=None, help="Extension to look for."
    )
    parser.add_argument(
        "--output", "-o", type=str, default=None, help="Name of output txt file."
    )

    parsed = parse_args(parser, verbose=False)

    if os.path.exists(parsed.directory):
        files = get_all_files_in_dir(
            parsed.directory,
            ext=parsed.extension,
            return_absolute=False,
            recursive=parsed.recursive,
        )
        output = (
            parsed.output
            if parsed.output is not None
            else os.path.join(parsed.directory, "current_contents.txt")
        )
        with open(output, "w") as f:
            for fname in files:
                f.write(f"{fname}\n")
    else:
        raise ValueError("Please pass a valid directory")


def cli_copy_files_in_dir():
    parser = argparse.ArgumentParser(description="Directory list command line")

    parser.add_argument(
        "input_directory", type=str, help="Directory to find files from.",
    )
    parser.add_argument(
        "output_directory", type=str, help="Directory to copy files to."
    )
    parser.add_argument(
        "--recursive",
        "-r",
        action="store_true",
        help="Whether to recurse into subdirectories.",
    )
    parser.add_argument(
        "--extension", "-e", type=str, default=None, help="Extension to look for."
    )
    parser.add_argument(
        "--regular_expression",
        "-re",
        type=str,
        default=None,
        help="Regular expression to filter files by.",
    )
    parser.add_argument(
        "--move", "-mv", action="store_true", help="Move files instead of copying them."
    )
    parser.add_argument(
        "--dummy",
        "-d",
        action="store_true",
        help="Dummy run, only print files that would be copied.",
    )

    parsed = parse_args(parser, verbose=False)

    if os.path.exists(parsed.input_directory):
        files = get_all_files_in_dir(
            parsed.input_directory,
            ext=parsed.extension,
            return_absolute=False,
            recursive=parsed.recursive,
            re_filter=parsed.regular_expression,
        )
        if parsed.dummy:
            name = "move" if parsed.move else "copy"
            print(f"Would {name} these files to {parsed.output_directory}:")
            for f in files:
                print(f)
            return

        os.makedirs(parsed.output_directory, exist_ok=True)
        for fname in files:
            output_loc = os.path.join(parsed.output_directory, fname)
            input_loc = os.path.join(parsed.input_directory, fname)
            if parsed.move:
                shutil.move(input_loc, output_loc)
            else:
                shutil.copy(input_loc, output_loc)
    else:
        raise ValueError("Please pass a valid directory")


if __name__ == "__main__":
    # cli_entry()
    cli_copy_files_in_dir()



"""PDF file related utilities."""

import os
from datetime import datetime
import argparse

from PyPDF2 import PdfFileMerger

from skm_pyutils.py_path import get_all_files_in_dir
from skm_pyutils.py_config import parse_args


def pdf_cat(input_files, output_location):
    """
    Concatenate PDF files.

    Parameters
    ----------
    input_files : list of str
        Paths to files to concatenate.
    output_location : str, optional
        The path to the output merged PDF location.

    Returns
    -------
    None

    """
    if output_location is None:
        now = datetime.now()
        # current_time = now.strftime("%H-%M-%S")
        whole_time = now.strftime("%Y-%m-%d--%H-%M-%S")
        output_location = "pdf_merge_" + whole_time + ".pdf"

    merger = PdfFileMerger()
    n_merged = 0
    for pdf in input_files:
        if os.path.exists(pdf):
            print(f"Merging {pdf}")
            merger.append(pdf)
            n_merged += 1
        else:
            print("Warning: {} does not exist".format(pdf))
    print(f"Saving merged output of {n_merged} files to {output_location}")
    merger.write(output_location)
    merger.close()


def merge_all_pdfs_in_dir(input_dir, out_name=None, recursive=False):
    """
    Merge all PDFs in the given directory.

    Parameters
    ----------
    input_dir : str
        The path to the directory.
    out_name : str, optional
        The name of the output file.
    recursive : bool, optional
        Whether to recurse through subdirectories. Defaults to False.
    
    Returns
    -------
    None

    """
    if out_name is None:
        now = datetime.now()
        # current_time = now.strftime("%H-%M-%S")
        whole_time = now.strftime("%Y-%m-%d--%H-%M-%S")
        out_name = "pdf_merge_" + whole_time + ".pdf"
        out_name = os.path.abspath(os.path.join(input_dir, out_name))

    pdf_files = get_all_files_in_dir(
        input_dir, ext=".pdf", recursive=recursive, return_absolute=True
    )

    pdf_cat(pdf_files, out_name)


def cli_entry():
    """Command line interface entry point."""
    parser = argparse.ArgumentParser(description="PDF merger command line")

    parser.add_argument(
        "--directory",
        "-d",
        type=str,
        default="",
        help="Directory to merge the pdf files from.",
    )
    parser.add_argument(
        "--recursive",
        "-r",
        action="store_true",
        help="Whether to recurse into subdirectories.",
    )
    parser.add_argument(
        "--item", "-i", action="append", type=str, help="Individual PDF files to merge."
    )
    parser.add_argument(
        "--output", "-o", type=str, default=None, help="Name of merged output PDF file."
    )

    parsed = parse_args(parser, verbose=False)

    if os.path.exists(parsed.directory):
        merge_all_pdfs_in_dir(
            parsed.directory, out_name=parsed.output, recursive=parsed.recursive
        )
    elif parsed.item is not None:
        pdf_cat(parsed.item, parsed.output)
    else:
        raise ValueError("Please pass items or a valid directory")


if __name__ == "__main__":
    cli_entry()



"""
This module handles printing items such as dicts.

One could alternatively use pprint for some of these.
"""
import rich


def walk_dict(d, depth=0):
    """Walk a Dictionary."""
    for k, v in sorted(d.items(), key=lambda x: x[0]):
        spaces = ("  ") * depth
        if hasattr(v, "items"):
            print(spaces + ("%s" % k))
            walk_dict(v, depth + 1)
        else:
            print(spaces + "%s %s" % (k, v))


def print_attrs(d, depth=0):
    """Walk through attributes and print them."""
    for k, v in sorted(d.items(), key=lambda x: x[0]):
        spaces = ("  ") * depth
        spaces2 = ("  ") * (depth + 1)
        if len(v.attrs) != 0:
            print("{}{} has attrs:".format(spaces, k))
            for key, val in v.attrs.items():
                print("{}{} {}".format(spaces2, key, val))
        if hasattr(v, "items"):
            print(spaces + ("%s" % k))
            print_attrs(v, depth=depth + 1)


def print_h5(file_location):
    """Print a summary of a h5 file."""
    import h5py

    with h5py.File(file_location, "r", libver="latest") as f:
        print("--------AVAILABLE METADATA--------")
        for key, val in f.attrs.items():
            print(key, val)
        print_attrs(f)
        print("\n--------AVAILABLE INFO--------")
        walk_dict(f)
        print()


inspect = rich.inspect



"""Array manipulation."""
import numpy as np


def non_zero_divide(numerator, denominator):
    """Divide numerator / denominator, but if denominator is zero result is zero"""
    result = np.divide(
        numerator, denominator, out=np.zeros_like(numerator), where=denominator != 0
    )
    return result



import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


DESCRIPTION = "skm_pyutils: Python Utilities"
LONG_DESCRIPTION = read("README.md")
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"

DISTNAME = "skm_pyutils"
MAINTAINER = "Sean Martin"
MAINTAINER_EMAIL = "martins7@tcd.ie"
URL = "https://github.com/seankmartin/PythonUtils"
DOWNLOAD_URL = "https://github.com/seankmartin/PythonUtils/archive/0.1.5.tar.gz"
VERSION = "0.1.5"

INSTALL_REQUIRES = ["numpy", "matplotlib", "seaborn", "PyPDF2", "rich", "psutil", "pyyaml"]

PACKAGES = ["skm_pyutils"]

CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
]

ENTRY_POINTS = {
    "console_scripts": [
        "pdf-merge = skm_pyutils.py_pdf:cli_entry",
        "dir-list = skm_pyutils.py_path:cli_entry",
        "copy-files = skm_pyutils.py_path:cli_copy_files_in_dir",
    ]
}

try:
    from setuptools import setup

    _has_setuptools = True
except ImportError:
    from distutils.core import setup

if __name__ == "__main__":

    setup(
        name=DISTNAME,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
        url=URL,
        version=VERSION,
        download_url=DOWNLOAD_URL,
        install_requires=INSTALL_REQUIRES,
        include_package_data=True,
        packages=PACKAGES,
        classifiers=CLASSIFIERS,
        entry_points=ENTRY_POINTS,
    )



"""
Logging related functions.
"""
import datetime
import logging
import os
import sys
import traceback

import psutil

from skm_pyutils.py_path import make_path_if_not_exists


def log_exception(ex, more_info="", location=None):
    """
    Log an expection to file and additional info.

    Parameters
    ----------
    ex : Exception
        The python exception that occurred
    more_info : str, optional
        Additional string to log, default is ""
    location : str, optional
        Where to store the log, default is
        home/.skm_python/caught_errors.txt

    Returns
    -------
    None

    """
    if location is None:
        default_loc = get_default_log_loc("caught_errors.txt")
    else:
        default_loc = location

    now = datetime.datetime.now()
    make_path_if_not_exists(default_loc)
    with open(default_loc, "a+") as f:
        f.write("\n----------Caught Exception at {}----------\n".format(now))
        traceback.print_exc(file=f)
    logging.error(
        "{} failed with caught exception.\nSee {} for more information.".format(
            more_info, default_loc
        ),
        exc_info=False,
    )


def default_excepthook(exc_type, exc_value, exc_traceback):
    """Any uncaught exceptions will be logged from here."""
    default_loc = get_default_log_loc("uncaught_errors.txt")

    file_logger = logging.getLogger(__name__)
    file_logger.propagate = False
    handler = logging.FileHandler(default_loc)
    file_logger.addHandler(handler)

    # Don't catch CTRL+C exceptions
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    now = datetime.datetime.now()
    file_logger.critical(
        "\n----------Uncaught Exception at {}----------".format(now),
        exc_info=(exc_type, exc_value, exc_traceback),
    )

    print("\nA fatal error occurred in this Python program")
    print(
        "The error info was: {}".format(
            "".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            ).strip()
        )
    )
    print("Please report this to {} and provide the file {}".format("us", default_loc))

    sys.exit(-1)


def override_excepthook(excepthook=None):
    """
    Change sys.excepthook behaviour.
    
    Parameters
    ----------
    excepthook : function, optional
        This is what sys.excepthook will be overridden by.
        See https://docs.python.org/3/library/sys.html#sys.excepthook
        for the function specification.
        If None is passed, uses the default excepthook in this module.
    
    Returns
    -------
    None

    """
    if excepthook is None:
        excepthook = default_excepthook
    sys.excepthook = excepthook


def get_default_log_loc(name):
    default_loc = os.path.join(os.path.expanduser("~"), ".skm_python", name)

    return default_loc


def setup_text_logging(
    in_dir, loglevel, bname="logfile.log", append=False, logname=None
):
    """
    Pass logging file location to logging.

    Parameters
    ----------
    in_dir : str
        Directory to save log file to. Can be None to use bname directly.
    loglevel : str
        The name of the log level to use, e.g. "debug" or 10
    bname : str, optional, defaults to "logfile.log"
        The basename of the log file to save to.
    append: bool, optional, defaults to False
    logname : str, optional, defaults to None
        If provided, creates this logger. Otherwise uses the root logger.

    Returns
    -------
    None

    """
    try:
        numeric_level = int(loglevel)
    except BaseException:
        numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % loglevel)

    if in_dir is None:
        fname = bname
    else:
        fname = os.path.join(in_dir, bname)
    if not append:
        if os.path.isfile(fname):
            open(fname, "w").close()

    root = logging.getLogger(logname)
    root.setLevel(numeric_level)
    file_handler = logging.FileHandler(fname)
    file_handler.setFormatter(
        logging.Formatter(
            fmt="%(levelname)s: %(asctime)s %(message)s",
            datefmt="%d/%m/%Y %I:%M:%S %p",
        )
    )
    root.addHandler(file_handler)
    # Encoding only supported in 3.9+
    # logging.basicConfig(
    #     filename=fname,
    #     # encoding="utf-8",
    #     level=numeric_level,
    #     format="%(levelname)s: %(asctime)s %(message)s",
    #     datefmt="%d/%m/%Y %I:%M:%S %p",
    # )
    mpl_logger = logging.getLogger("matplotlib")
    mpl_logger.setLevel(level=logging.WARNING)

    # print("See {} for {} level logs".format(fname, loglevel))


class FileStdoutLogger:
    """A logger that prints to stdout and to a file."""

    def __init__(self, name="stdout_logger"):
        self.name = name
        self.logger = None
        self.create_logger()

    def init_logging(self):
        self.logger.setLevel(logging.INFO)
        out_loc = self.get_default_log_location()
        output_file_handler = logging.FileHandler(out_loc)
        stdout_handler = logging.StreamHandler(sys.stdout)

        self.logger.addHandler(output_file_handler)
        self.logger.addHandler(stdout_handler)

    def create_logger(self):
        if self.logger is None:
            self.logger = logging.getLogger(self.name)
        if len(self.get_handlers()) == 0:
            self.init_logging()

    def print(self, msg):
        self.logger.info(msg)

    def get_default_log_location(self):
        home = os.path.expanduser("~")
        out_loc = os.path.join(home, ".skm_python", f"{self.name}.log")
        os.makedirs(os.path.dirname(out_loc), exist_ok=True)

        return out_loc

    def read_log_file(self):
        loc = self.get_default_log_location()
        if os.path.exists(loc):
            with open(loc, "r") as f:
                out = f.read().strip()
            return out
        else:
            return "No log file exists."

    def clear_log_file(self):
        # handlers = self.logger.handlers[:]
        # for handler in handlers:
        #     handler.close()
        #     self.logger.removeHandler(handler)

        loc = self.get_default_log_location()
        if os.path.exists(loc):
            open(loc, "w").close()

    def get_handlers(self):
        return self.logger.handlers[:]


class FileLogger:
    """A logger that prints to a file."""

    def __init__(self, name="main", level="warning"):
        self.name = name
        self.logger = None
        self.level = level
        self.create_logger()

    def init_logging(self):
        out_loc = self.get_default_log_location()
        setup_text_logging(None, self.level, out_loc, append=True, logname=self.name)

    def create_logger(self):
        if self.logger is None:
            self.logger = logging.getLogger(self.name)
        if len(self.get_handlers()) == 0:
            self.init_logging()

    def info(self, msg, **kwargs):
        self.logger.info(msg, **kwargs)

    def warning(self, msg, **kwargs):
        self.logger.warning(msg, **kwargs)

    def error(self, msg, **kwargs):
        self.logger.error(msg, **kwargs)

    def critical(self, msg, **kwargs):
        self.logger.critical(msg, **kwargs)

    def debug(self, msg, **kwargs):
        self.logger.debug(msg, **kwargs)

    def set_level(self, loglevel):
        try:
            numeric_level = int(loglevel)
        except BaseException:
            numeric_level = getattr(logging, loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError("Invalid log level: %s" % loglevel)

        self.logger.setLevel(numeric_level)
        filename = self.get_default_log_location()
        print("See {} for {} level logs".format(filename, loglevel))

    def get_default_log_location(self):
        home = os.path.expanduser("~")
        out_loc = os.path.join(home, ".skm_python", f"{self.name}.log")
        os.makedirs(os.path.dirname(out_loc), exist_ok=True)

        return out_loc

    def read_log_file(self):
        loc = self.get_default_log_location()
        if os.path.exists(loc):
            with open(loc, "r") as f:
                out = f.read().strip()
            return out
        else:
            return "No log file exists."

    def clear_log_file(self):
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)

        loc = self.get_default_log_location()
        if os.path.exists(loc):
            os.remove(loc)

    def get_handlers(self):
        return self.logger.handlers[:]


def print_memory_usage(as_string: bool=False) -> str:
    """
    Print memory usage information
    
    Parameters
    ----------
    as_string : bool, optional
        Just return the string representation, don't print.
        By default, False.
    
    Returns
    -------
    usage : str
        String about memory usage
    """
    str_ = f"RAM memory usage stats: {psutil.virtual_memory()}"
    if not as_string:
        print(str_)
    return str_



"""Config file and logging related utility functions."""
import os
import sys
import configparser
import yaml
import json
from pprint import pprint


def read_cfg(location, verbose=True):
    """
    Read config file at location using ConfigParser.

    Parameters
    ----------
    location : str
        Where the config file is located
    verbose : bool, optional, defaults to True
        Should print the contents of the read config file.

    Returns
    -------
    ConfigParser
        The python ConfigParser object after reading the cfg.

    """
    if not os.path.exists(location):
        raise ValueError(f"Config file {location} does not exist")

    config = configparser.ConfigParser()
    config.read(location)

    if verbose:
        print_cfg(config, "Program started with configuration")
    return config


def print_cfg(config, msg=""):
    """
    Print the contents of a ConfigParser object.

    Parameters
    ----------
    config : ConfigParser
        The ConfigParser to print the contents of.
    msg: str, optional, defaults to ""
        Message to print before printing the config file.

    Returns
    -------
    None

    """
    if msg != "":
        print(msg)
    config_dict = [{x: tuple(config.items(x))} for x in config.sections()]
    pprint(config_dict, width=120)


def parse_args(parser, verbose=True):
    """
    Parse command line arguments into a Namespace.

    Parameters
    ----------
    verbose : bool, optional, defaults to True
        Should print the values of the command line args.

    Returns
    -------
    Namespace
        Parsed arguments.

    Raises
    ------
    ValueError
        If any arguments are passed which are not used in program.

    """
    args, unparsed = parser.parse_known_args()

    if len(unparsed) != 0:
        raise ValueError(
            "Unrecognised command line arguments passed {}".format(unparsed)
        )

    if verbose:
        if len(sys.argv) > 1:
            print("Command line arguments", args)
    return args


def read_python(path, dirname_replacement=""):
    """
    Execute a python script at path.

    The script is expected to have items visible at global scope,
    which are stored as metadata.

    Note
    ----
    The string "__dirname__" is magic and will be replaced by the
    absolute path to the directory containing the script.

    Parameters
    ----------
    path : string
        The location of the python script.
    dirname_replacement : string, optional, optional, defaults to None
        What to replace __dirname__ with.
        By default, None will replace __dirname__ with dirname of path.

    Returns
    -------
    dict
        The scripts global scope variables stored in a dictionary.

    """

    def normalise_path(pth):
        s = os.path.abspath(pth)
        s = s.replace(os.sep, "/")
        return s

    path = os.path.realpath(os.path.expanduser(path))
    if not os.path.exists(path):
        raise ValueError("{} does not exist to read".format(path))
    with open(path, "r") as f:
        contents = f.read()
    if dirname_replacement != "":
        contents = contents.replace("__dirname__", normalise_path(dirname_replacement))
    else:
        contents = contents.replace(
            "__dirname__", normalise_path(os.path.dirname(path))
        )
    contents = contents.replace(
        "__thisdirname__", normalise_path(os.path.dirname(path))
    )
    metadata = {}
    try:
        exec(contents, {}, metadata)
    except Exception as e:
        import traceback
        print("QUITTING: An error occurred reading {}".format(path))
        traceback.print_exc()
        exit(-1)
    metadata = {k.lower(): v for (k, v) in metadata.items()}
    return metadata

def read_yaml(path):
    with open(path, 'r') as stream:
        parsed_yaml=yaml.safe_load(stream)
    return parsed_yaml


def read_json(path):
    with open(path, 'r') as stream:
        parsed_json = json.load(stream)
    return parsed_json

def split_dict(in_dict, index):
    """
    Grab the value at index from each list in the dictionary.

    Parameters
    ----------
    in_dict : dict
        The dictionary to grab from
    index : int
        The index in the lists to pull from

    Returns
    -------
    dict
        The original dictionary but with index values pulled out.

    """
    new_dict = {}
    for key, value in in_dict.items():
        if isinstance(value, list):
            new_dict[key] = value[index]
    return new_dict
